@@ -536,25 +536,6 @@ quantile2.rvar <- function(
 
 # internal ----------------------------------------------------------------
 
-#' Find the optimal next size for the FFT so that a minimum number of zeros
-#' are padded.
-#' @param N length of the sequence over which to apply FFT
-#' @return the optimal next step size as a single integer
-#' @noRd
-fft_next_good_size <- function(N) {
-  if (N <= 2)
-    return(2)
-  while (TRUE) {
-    m <- N
-    while ((m %% 2) == 0) m <- m / 2
-    while ((m %% 3) == 0) m <- m / 3
-    while ((m %% 5) == 0) m <- m / 5
-    if (m <= 1)
-      return(N)
-    N <- N + 1
-  }
-}
-
 #' Autocovariance estimates
 #'
 #' Compute autocovariance estimates for every lag for the specified
@@ -567,8 +548,13 @@ fft_next_good_size <- function(N) {
 #' @export
 autocovariance <- function(x) {
   N <- length(x)
+  varx <- var(x)
+  if (varx==0) {
+    # if variance is 0, then all autocovariances are 0
+    return(rep(0, N))
+  }
   # zero padding makes fft() much faster when N > 1000
-  M <- fft_next_good_size(N)
+  M <- nextn(N)
   Mt2 <- 2 * M
   yc <- x - mean(x)
   yc <- c(yc, rep.int(0, Mt2 - N))
@@ -576,7 +562,7 @@ autocovariance <- function(x) {
   ac <- Re(fft(abs(fft(yc))^2, inverse = TRUE)[1:N])
   # use "biased" estimate as recommended by Geyer (1992)
   # direct scaling with var(x) avoids need to compute "mask effect"
-  ac <- ac / ac[1] * var(x) * (N - 1) / N
+  ac <- ac / ac[1] * varx * (N - 1) / N
   ac
 }
 
@@ -694,12 +680,8 @@ fold_draws <- function(x) {
   }
   nchains <- NCOL(x)
   niterations <- NROW(x)
-  chain_mean <- numeric(nchains)
-  chain_var <- numeric(nchains)
-  for (i in seq_len(nchains)) {
-    chain_mean[i] <- mean(x[, i])
-    chain_var[i] <- var(x[, i])
-  }
+  chain_mean <- matrixStats::colMeans2(x)
+  chain_var <- matrixStats::colVars(x, center=chain_mean)
   var_between <- niterations * var(chain_mean)
   var_within <- mean(chain_var)
   sqrt((var_between / var_within + niterations - 1) / niterations)
@@ -716,14 +698,12 @@ fold_draws <- function(x) {
   if (niterations < 3L || should_return_NA(x)) {
     return(NA_real_)
   }
-  acov_fun <- function(i) autocovariance(x[, i])
-  acov <- lapply(seq_len(nchains), acov_fun)
-  acov <- do.call(cbind, acov)
-  chain_mean <- apply(x, 2, mean)
-  mean_var <- mean(acov[1, ]) * niterations / (niterations - 1)
+  acov <- apply(x, 2, autocovariance)
+  acov_means <- matrixStats::rowMeans2(acov)
+  mean_var <- acov_means[1] * niterations / (niterations - 1)
   var_plus <- mean_var * (niterations - 1) / niterations
   if (nchains > 1) {
-    var_plus <- var_plus + var(chain_mean)
+    var_plus <- var_plus + var(matrixStats::colMeans2(x))
   }
 
   # Geyer's initial positive sequence
@@ -731,13 +711,13 @@ fold_draws <- function(x) {
   t <- 0
   rho_hat_even <- 1
   rho_hat_t[t + 1] <- rho_hat_even
-  rho_hat_odd <- 1 - (mean_var - mean(acov[t + 2, ])) / var_plus
+  rho_hat_odd <- 1 - (mean_var - acov_means[t + 2]) / var_plus
   rho_hat_t[t + 2] <- rho_hat_odd
   while (t < NROW(acov) - 5 && !is.nan(rho_hat_even + rho_hat_odd) &&
          (rho_hat_even + rho_hat_odd > 0)) {
     t <- t + 2
-    rho_hat_even = 1 - (mean_var - mean(acov[t + 1, ])) / var_plus
-    rho_hat_odd = 1 - (mean_var - mean(acov[t + 2, ])) / var_plus
+    rho_hat_even = 1 - (mean_var - acov_means[t + 1]) / var_plus
+    rho_hat_odd = 1 - (mean_var - acov_means[t + 2]) / var_plus
     if ((rho_hat_even + rho_hat_odd) >= 0) {
       rho_hat_t[t + 1] <- rho_hat_even
       rho_hat_t[t + 2] <- rho_hat_odd
@@ -770,7 +750,21 @@ fold_draws <- function(x) {
 }
 
 # should NA be returned by a convergence diagnostic?
-should_return_NA <- function(x) {
-  anyNA(x) || checkmate::anyInfinite(x) || is_constant(x) ||
-    (is.matrix(x) && any(apply(x, 2, is_constant)))
+should_return_NA <- function(x, tol = .Machine$double.eps) {
+  if (anyNA(x) || checkmate::anyInfinite(x)) {
+    return(TRUE)
+  }
+  # checking for constant input per chain is too conservative for ess_tail
+  # so we don't check this for the time being until we find a better solution
+  # if (is.matrix(x)) {
+  #   # is_constant() vectorized over x columns
+  #   if (is.logical(x)) {
+  #     return(any(matrixStats::colAnys(x) == matrixStats::colAlls(x)))
+  #   } else if (is.integer(x)) {
+  #     return(any(matrixStats::rowDiffs(matrixStats::colRanges(x)) == 0L))
+  #   } else {
+  #     return(any(abs(matrixStats::rowDiffs(matrixStats::colRanges(x))) < tol))
+  #   }
+  # }
+  is_constant(x, tol = tol)
 }
